#include <stdio.h>
#include <stdlib.h>

#ifdef DICT_HW_MEASURE_TIMINGS
# include <sys/time.h>
#endif /* DICT_HW_MEASURE_TIMINGS */

#include "dict_hw.h"

#include "dict_image.h"

#include "helpers.h"

#include "local_sum.cu.h"

#include "normxcorr_hw.h"
#include "normxcorr_hw.cu.h"


#define CHECK_ERROR(err) \
    for (icp = 0; icp < ncp; icp += per_device) { \
	int i = icp / per_device; \
        ProcessingState ps = ctx->pstates[i]; \
	if (ps->status) { \
	    err = ps->status; \
	    break; \
	} \
    } \



static int device_number = 0;
#ifdef DICT_SUPPORT_THREADS
static int devices[DICT_HW_MAX_DEVICES];
#endif /* DICT_SUPPORT_THREADS */
static size_t device_memory = 2147483648;

#ifdef DICT_HW_MEASURE_TIMINGS
static int hardware_detection_time = 0;
#endif /* DICT_HW_MEASURE_TIMINGS */

int dictDetectHardware() {
    int memory;
    int deviceCount;
    cudaDeviceProp deviceProp;

    if (device_number) return device_number;
    
#ifdef DICT_HW_MEASURE_TIMINGS
    struct timeval tv1, tv2;
    gettimeofday(&tv1, NULL);
#endif /* DICT_HW_MEASURE_TIMINGS */


    cudaGetDeviceCount(&deviceCount);
    if (!deviceCount) return -1;

    for (int i = 0; i < deviceCount; i++) {
        cudaGetDeviceProperties(&deviceProp, i);
        if ((deviceProp.major > 1)||((deviceProp.major == 1)&&(deviceProp.minor > 2))) {
	    memory = deviceProp.totalGlobalMem;
	    if (memory >= CUDA_MIN_MEMORY) {
		if (memory < device_memory) device_memory = memory;
#ifdef DICT_SUPPORT_THREADS
    		devices[device_number++] = i;
		if (device_number == DICT_HW_MAX_DEVICES) break;
#else /* DICT_SUPPORT_THREADS */
		device_number = 1;
#endif /* DICT_SUPPORT_THREADS */
	    }
        }
    }

#ifdef DICT_HW_MEASURE_TIMINGS
    gettimeofday(&tv2, NULL);
    hardware_detection_time = (tv2.tv_sec - tv1.tv_sec) * 1000000 + (tv2.tv_usec - tv1.tv_usec);
#endif /* DICT_HW_MEASURE_TIMINGS */

    return device_number;
}


#ifdef DICT_SUPPORT_THREADS
static int dictFreeThread(HWThread thr, void *hwctx, int device_id, void *data) {
    DICTContext ctx = (DICTContext)data;

    ProcessingState ps = ctx->pstates[device_id];
    if (ps) {
# ifdef SEQUENTIAL_CUDPP_DESTROY	
	hw_sched_lock(ctx->sched, data);
# endif SEQUENTIAL_CUDPP_DESTROY	
	fftFree(ps);
# ifdef SEQUENTIAL_CUDPP_DESTROY	
	hw_sched_unlock(ctx->sched, data);
# endif SEQUENTIAL_CUDPP_DESTROY	
    }
            
    return 0;
}
#endif /* DICT_SUPPORT_THREADS */

static void dictFree(DICTContext ctx) {
    int i;
    
    if (ctx->pstates) {
	if (ctx->use_threads) {
#ifdef DICT_SUPPORT_THREADS 
	    hw_sched_schedule_task(ctx->sched, ctx, dictFreeThread);
	    hw_sched_wait_task(ctx->sched);
#endif /* DICT_SUPPORT_THREADS */
        } else {
    	    for (i = 0; i < device_number; i++) {
		ProcessingState ps = ctx->pstates[i];
		if (ps) fftFree(ps);
	    }
	}
    }

    if (ctx->image_buf) {
	dictImageFree(ctx);
    }
    
    if (ctx->preload_image) free(ctx->preload_image);
    if (ctx->preload_request) free(ctx->preload_request);
    if (ctx->base_image) free(ctx->base_image);
    if (ctx->points) cudaFreeHost(ctx->points);

    memset(ctx, 0, ((char*)&(ctx->matlab_mode)) - ((char*)ctx));
}


size_t dictGetMaxCacheablePoints(int corr_size, int precision, int flags) {
    // we are a bit pessimistic here

    int fft_size = 6 * corr_size + 1;
    int fft_real_size;
    
    if (flags&DICT_FLAGS_FIXED_FFT_SIZE) {
        fft_real_size = fft_size;
    } else {
        fft_real_size = next_power(fft_size);
    }

    int fft_alloc_size = calc_alloc(fft_real_size * fft_real_size, BLOCK_SIZE_1D);
    int side_alloc_size = calc_alloc(fft_size, SIDE_BLOCK_SIZE);

    int subimage_size = corr_size * 4 + 1;
    int lsum_size = corr_size * 2 + 1;
    int lsum_temp_size = subimage_size + 2 * lsum_size - 1;

    int lsum_alloc_size = calc_alloc(lsum_temp_size + lsum_size, BLOCK_SIZE_2D);

    int size = max3(
	((1 + CP_BLOCK) * fft_alloc_size) * sizeof(cufftComplex),		/* FFT multiplication */
	2 * CP_BLOCK * side_alloc_size * sizeof(int32_t),			/* Sum, Std computations */
	CP_BLOCK * side_alloc_size * (sizeof(int32_t) + sizeof(float))		/* Max of correlation */
    );

    int ncp = (device_memory - CUDA_EXTRA_MEMORY - (
	    size + 								// cuda_temp_buffer
	    CP_BLOCK * side_alloc_size * side_alloc_size * sizeof(uint8_t) +	// cuda_input_buffer
	    2 * CP_BLOCK * fft_alloc_size * sizeof(cufftReal) +			// cuda_base_buffer, cuda_data_buffer
	    4 * lsum_alloc_size * lsum_alloc_size * sizeof(float)		// cuda_lsum_temp
	)) / (
	    2 * sizeof(float) +							// cuda_points
	    fft_alloc_size * (2 * sizeof(float) + sizeof(cufftComplex))		// caches
	);

    int	ncp_blocks = ncp / CP_BLOCK;						// thats per device

    return (flags&DICT_FLAGS_SINGLE_THREAD)?(CP_BLOCK * ncp_blocks):(CP_BLOCK * device_number * ncp_blocks);
}

#ifdef DICT_SUPPORT_THREADS
static int dictInitThread(HWThread thr, void *hwctx, int device_id, void *data) {
    cudaError cuda_err;

    cuda_err = cudaSetDevice(devices[device_id]);
    if (cuda_err) {
	reportError("Failed to associate cuda device");
    }
        
    return 0;
}
#endif /* DICT_SUPPORT_THREADS */

DICTContext dictCreateContext() {
    int i;
    
#ifdef DICT_HW_MEASURE_TIMINGS
    struct timeval tv1, tv2;
#endif /* DICT_HW_MEASURE_TIMINGS */

    if (!device_number) {
        if (dictDetectHardware() <= 0) return NULL;
    }

#ifdef DICT_HW_MEASURE_TIMINGS
    gettimeofday(&tv1, NULL);
#endif /* DICT_HW_MEASURE_TIMINGS */
    
    DICTContext ctx = (DICTContext)malloc(sizeof(TDICTContext));
    if (!ctx) return ctx;

    memset(ctx, 0, sizeof(TDICTContext));
    
    ProcessingState *pstates = (ProcessingState*)malloc(device_number * sizeof(ProcessingState));
    if (!pstates) {
	dictDestroyContext(ctx);
	return NULL;
    }
    
    memset(pstates, 0, device_number * sizeof(ProcessingState));
    ctx->pstates = pstates;

    for (i = 0; i < device_number; i++) {
	TProcessingState *pstate = pstateInit();
	if (!pstate) {
	    dictDestroyContext(ctx);
	    return NULL;
	}
	
	ctx->pstates[i] = pstate;
    }

#ifdef DICT_SUPPORT_THREADS
    ctx->sched = hw_sched_create(device_number);
    if (!ctx->sched) {
	dictDestroyContext(ctx);
	return NULL;
    }

    hw_sched_schedule_task(ctx->sched, ctx, dictInitThread);
    hw_sched_wait_task(ctx->sched);
#endif /* DICT_SUPPORT_THREADS */

#ifdef DICT_HW_MEASURE_TIMINGS
    gettimeofday(&tv2, NULL);
    ctx->time[0] = hardware_detection_time;
    ctx->time[1] = (tv2.tv_sec - tv1.tv_sec) * 1000000 + (tv2.tv_usec - tv1.tv_usec);
#endif /* DICT_HW_MEASURE_TIMINGS */

    return ctx;
}

void dictDestroyContext(DICTContext ctx) {
    int i;
    
#ifdef DICT_HW_MEASURE_TIMINGS
    int time;
    struct timeval tv1, tv2;

    int init_time = ctx->time[0]+ctx->time[1]+ctx->time[2]+ctx->time[3]+ctx->time[12]+ctx->time[13]+ctx->time[17];
    int load_time = ctx->time[14] + ctx->time[15];
    int process_time = ctx->time[5];
    
    reportMessage("");
    reportMessage("Library timings");
    reportMessage("---------------");
    print_timing("Overall", init_time + load_time + process_time + ctx->time[16]);
    print_timing("Initialization", init_time);
    print_timing("  Hardware Detection", ctx->time[0]);
    print_timing("  Context Initialization", ctx->time[1]);
    print_timing("  Setting Template Points", ctx->time[2]);
    print_timing("  Setting Current Points", ctx->time[3]);
    print_timing("  Loading Template Image", ctx->time[12]);
    print_timing("  Reducting Template Image", ctx->time[13]);
    print_timing("  Processing Template Image", ctx->time[17]);

    print_timing("Loading Images", load_time);
    print_timing("  Load", ctx->time[14]);
    print_timing("  Reduction", ctx->time[15]);
    print_timing("Processing Images", process_time);

    if (!ctx->use_threads) {
	print_timing("  Processing Template Image", ctx->time[6]);
	print_timing("  Copy Image   ", ctx->time[4]);
	print_timing("  Load Fragment", ctx->time[7]);
	print_timing("  Preprocessing", ctx->time[8]);
	print_timing("  FFT", ctx->time[9]);
	print_timing("  Postprocessing", ctx->time[10]);
    }
    print_timing("Waiting Result", ctx->time[16]);

    gettimeofday(&tv1, NULL);
#endif /* DICT_HW_MEASURE_TIMINGS */

    dictFree(ctx);

#ifdef DICT_SUPPORT_THREADS
    if (ctx->sched) hw_sched_destroy(ctx->sched);
#endif /* DICT_SUPPORT_THREADS */

    if (ctx->pstates) {
	for (i = device_number; i > 0; --i) {
	    if (ctx->pstates[i - 1]) pstateFree(ctx->pstates[i - 1]);
	}
	free(ctx->pstates);
    }
    free(ctx);

#ifdef DICT_HW_MEASURE_TIMINGS
    gettimeofday(&tv2, NULL);
    time = (tv2.tv_sec - tv1.tv_sec) * 1000000 + (tv2.tv_usec - tv1.tv_usec);
    print_timing("Destruction", time);
    reportMessage("");
#endif /* DICT_HW_MEASURE_TIMINGS */
}

int dictSetLogger(DICTLogger error_reporter, DICTLogger message_writer) {
    return reportSet(error_reporter, message_writer);
}

static int dictSetupFFT(DICTContext ctx, ProcessingState ps, int ncp, int corr_size, int precision, int flags) {
    int err = 0;
    int base_blocks, side_blocks;

    ps->ncp = ncp;
    ps->corr_size = corr_size;
    ps->precision = precision;

    ps->subimage_size = ps->corr_size * 4 + 1;
    ps->fft_size = 6 * corr_size + 1;

    if (flags&DICT_FLAGS_FIXED_FFT_SIZE) {
        ps->fft_real_size = ps->fft_size;
    } else {
        ps->fft_real_size = next_power(ps->fft_size);
    }
    
    if (flags&DICT_FLAGS_MATLAB_MODE) {
	ps->matlab_mode = 1;
    } else {
	ps->matlab_mode = 0;
    }

    ps->ncp_alloc_size = calc_alloc(ps->ncp, CP_BLOCK);
    ps->side_alloc_size = calc_alloc(ps->fft_size, SIDE_BLOCK_SIZE);

    ps->fft_alloc_size = calc_alloc(ps->fft_real_size * ps->fft_real_size, BLOCK_SIZE_1D);

    ps->lsum_size = ps->corr_size * 2 + 1;
    ps->lsum_temp_size = ps->subimage_size + 2*ps->lsum_size - 1;

    ps->lsum_short_aligned_size = calc_alloc(ps->fft_size, BLOCK_SIZE_2D);
    ps->lsum_aligned_size = calc_alloc(ps->lsum_temp_size, BLOCK_SIZE_2D);
    ps->lsum_alloc_size = calc_alloc(ps->lsum_temp_size + ps->lsum_size, BLOCK_SIZE_2D);

    side_blocks = calc_blocks(2 * ps->corr_size  + 1, SIDE_BLOCK_SIZE);
    ps->side_blocks_power = get_power(side_blocks);

    base_blocks = calc_blocks(4 * ps->corr_size  + 1, BLOCK_SIZE_1D);
    ps->base_blocks_power = get_power(base_blocks);

    ps->use_cache = 1;

    if (!ctx->use_threads) {
	err = fftInit(ps, device_memory);
	if ((err == DICT_ERROR_CUDA_MALLOC)&&(ps->use_cache)) {
	    cudaGetLastError();
	    ps->use_cache = 0;
	    err = fftInit(ps, device_memory);
	}
    }

    return err;
}

#ifdef DICT_SUPPORT_THREADS
static int dictSetupThread(HWThread thr, void *hwctx, int device_id, void *data) {
    int err;
    DICTContext ctx = (DICTContext)data;
    
    ProcessingState ps = ctx->pstates[device_id];

    err = fftInit(ps, device_memory);
    if ((err == DICT_ERROR_CUDA_MALLOC)&&(ps->use_cache)) {
	cudaGetLastError();
	ps->use_cache = 0;
	err = fftInit(ps, device_memory);
    }

    ps->status = err;
    
    return 0;
}
#endif /* DICT_SUPPORT_THREADS */


int dictSetup(DICTContext ctx, int ncp, int corr_size, int precision, int flags) {
    int i, icp;
    
    int err;
    cudaError cuda_err;

    int have_cache = 0;

    int ncp_alloc = calc_alloc(ncp, CP_BLOCK);
    int ncp_blocks;
    int blocks_alloc;
    int per_device;

#ifdef DICT_HW_MEASURE_TIMINGS
    struct timeval tv1, tv2;
    gettimeofday(&tv1, NULL);
#endif /* DICT_HW_MEASURE_TIMINGS */

    dictFree(ctx);
    
    if (flags&DICT_FLAGS_MATLAB_MODE) {
	ctx->matlab_mode = 1;
    } else {
	ctx->matlab_mode = 0;
    }

#ifdef DICT_SUPPORT_THREADS 
    if (flags&DICT_FLAGS_SINGLE_THREAD) {
	ctx->use_threads = 0;
    } else {
	ctx->use_threads = 1;
    }
#endif /* DICT_SUPPORT_THREADS */

    if (ctx->use_threads) {
	ncp_blocks = ncp_alloc / CP_BLOCK;
	blocks_alloc = calc_alloc(ncp_blocks, device_number);
	per_device = (blocks_alloc / device_number) * CP_BLOCK;
    } else {
	per_device = ncp_alloc;
    }

    ctx->ncp = ncp;
    ctx->ncp_alloc_size = ncp_alloc;
    ctx->ncp_per_device = per_device;

    cuda_err = cudaHostAlloc((void**)&ctx->points, 8 * ncp_alloc * sizeof(float), 0);
    if (cuda_err) {
	reportError("Page locked host memory allocation of 8*%u*float bytes for control points is failed", ncp_alloc);
	dictFree(ctx);
	return DICT_ERROR_MALLOC;
    }
    
    for (icp = 0; icp < ncp; icp += per_device) {
	i = icp / per_device;
	int device_ncp = min2(per_device, ncp - icp);
        ProcessingState ps = ctx->pstates[i];

	err = dictSetupFFT(ctx, ps, device_ncp, corr_size, precision, flags);
	if (err) return err;
    }
    
#ifdef DICT_SUPPORT_THREADS 
    if ((i + 1) < device_number) hw_sched_limit_num_threads(ctx->sched, i + 1);
    else hw_sched_limit_num_threads(ctx->sched, 0);
    
    if (ctx->use_threads) {
	hw_sched_schedule_task(ctx->sched, ctx, dictSetupThread);
	hw_sched_wait_task(ctx->sched);
    }
#endif /* DICT_SUPPORT_THREADS */
    
    for (icp = 0; icp < ncp; icp += per_device) {
	i = icp / per_device;
        ProcessingState ps = ctx->pstates[i];

	if (ps->status) return ps->status;

	ps->points = ctx->points + icp;
	ps->points_alloc_size = ncp_alloc;

	if (ps->use_cache) ++have_cache;
    }

#ifdef DICT_HW_MEASURE_TIMINGS
    gettimeofday(&tv2, NULL);
    ctx->time[1] += (tv2.tv_sec - tv1.tv_sec) * 1000000 + (tv2.tv_usec - tv1.tv_usec);
#endif /* DICT_HW_MEASURE_TIMINGS */

    if (have_cache) {
	reportMessage("Caching is enabled for %i devices of %i", have_cache, i + 1);
    } else {
    	reportMessage("Caching is disabled for all %i devices", i + 1);
    }
    
    ctx->configured_ncp = ctx->ncp;

    return 0;
}

int dictSetActualNumberOfPoints(DICTContext ctx, int ncp) {
    int i, icp;
    
    if (ncp > ctx->configured_ncp) {
	reportMessage("DICT Context is configured for maximum %i control points, but %i supplied", ctx->configured_ncp, ncp);
	return DICT_ERROR_PARAM;
    }
    
    int ncp_alloc = calc_alloc(ncp, CP_BLOCK);
    int ncp_blocks;
    int blocks_alloc;
    int per_device;

    if (ctx->use_threads) {
	ncp_blocks = ncp_alloc / CP_BLOCK;
	blocks_alloc = calc_alloc(ncp_blocks, device_number);
	per_device = (blocks_alloc / device_number) * CP_BLOCK;
    } else {
	per_device = ncp_alloc;
    }

    ctx->ncp = ncp;
    ctx->ncp_alloc_size = ncp_alloc;
    ctx->ncp_per_device = per_device;

    for (icp = 0; icp < ncp; icp += per_device) {
	i = icp / per_device;
        ProcessingState ps = ctx->pstates[i];

	ps->ncp = min2(per_device, ncp - icp);
	ps->points = ctx->points + icp;
	ps->points_alloc_size = ncp_alloc;
    }


    return 0;
}

int dictClean(DICTContext ctx) {
    int icp, ncp = ctx->ncp;
    int per_device = ctx->ncp_per_device;
    
    for (icp = 0; icp < ncp; icp += per_device) {
	int i = icp / per_device;
	ProcessingState ps = ctx->pstates[i];
	memset(ps->banlist, 1, ps->ncp * sizeof(uint8_t));
    }
    
    return 0;
}

#ifdef DICT_SUPPORT_THREADS 
static int dictSetDimensionsThread(HWThread thr, void *hwctx, int device_id, void *data) {
    int err;
    DICTContext ctx = (DICTContext)data;
    ProcessingState ps = ctx->pstates[device_id];

    err = fftSetupDimensions(ps);
    ps->status = err;
    
    return 0;
}
#endif /* DICT_SUPPORT_THREADS */

int dictSetDimensions(DICTContext ctx, int width, int height) {
    int err = 0;
    
    int i;

    if ((width != ctx->width)||(height != ctx->height)) {
	int image_size = width * height;
	
        if (ctx->base_image) free(ctx->base_image);
	ctx->base_image = (unsigned char*)malloc(image_size * sizeof(uint8_t));
	if (!ctx->base_image) {
	    reportError("Memory allocation of %u*%u*uint8_t bytes for DICT/base_image is failed", width, height);
	    return DICT_ERROR_MALLOC;
	}

	ctx->width = width;
	ctx->height = height;

	for (i = 0; i < device_number; i++) {
	    ProcessingState ps = ctx->pstates[i];
	    
	    ps->width = width;
	    ps->height = height;

	    if (!ctx->use_threads) {
		err = fftSetupDimensions(ps);
	        if (err) return err;
	    }
	}

#ifdef DICT_SUPPORT_THREADS 
	if (ctx->use_threads) {
	    hw_sched_schedule_task(ctx->sched, ctx, dictSetDimensionsThread);
	    hw_sched_wait_task(ctx->sched);

	    for (i = 0; i < device_number; i++) {
		ProcessingState ps = ctx->pstates[i];
		if (ps->status) {
		    err = ps->status;
	    	    break;
		}
	    }
	}
#endif /* DICT_SUPPORT_THREADS */
    }    

    return err;
}



int dictSetTemplatePoints(DICTContext ctx, const float *points_x, const float *points_y) {
#ifdef DICT_HW_MEASURE_TIMINGS
    struct timeval tv1, tv2;
    gettimeofday(&tv1, NULL);
#endif /* DICT_HW_MEASURE_TIMINGS */

    memcpy(ctx->points,                       points_x, ctx->ncp * sizeof(float));
    memcpy(ctx->points + ctx->ncp_alloc_size, points_y, ctx->ncp * sizeof(float));

#ifdef DICT_HW_MEASURE_TIMINGS
    gettimeofday(&tv2, NULL);
    ctx->time[2] += (tv2.tv_sec - tv1.tv_sec) * 1000000 + (tv2.tv_usec - tv1.tv_usec);
#endif /* DICT_HW_MEASURE_TIMINGS */

    return 0;
}


int dictSetPointsBuffer(DICTContext ctx, float *point_x, float *point_y) {
    int icp, ncp = ctx->ncp;
    int per_device = ctx->ncp_per_device;
    
    ctx->res_x = point_x;
    ctx->res_y = point_y;
    
    for (icp = 0; icp < ncp; icp += per_device) {
	int i = icp / per_device;
        ProcessingState ps = ctx->pstates[i];

	ps->res_x = point_x + icp;
	ps->res_y = point_y + icp;
    }
    
    return 0;
}

int dictSetCurrentPoints(DICTContext ctx, const float *points_x, const float *points_y) {
    int icp, ncp = ctx->ncp;
    int per_device = ctx->ncp_per_device;

#ifdef DICT_HW_MEASURE_TIMINGS
    struct timeval tv1, tv2;
    gettimeofday(&tv1, NULL);
#endif /* DICT_HW_MEASURE_TIMINGS */

    memcpy(ctx->points + 2 * ctx->ncp_alloc_size, points_x, ctx->ncp * sizeof(float));
    memcpy(ctx->points + 3 * ctx->ncp_alloc_size, points_y, ctx->ncp * sizeof(float));

#ifdef DICT_HW_MEASURE_TIMINGS
    gettimeofday(&tv2, NULL);
    ctx->time[3] += (tv2.tv_sec - tv1.tv_sec) * 1000000 + (tv2.tv_usec - tv1.tv_usec);
#endif /* DICT_HW_MEASURE_TIMINGS */

    for (icp = 0; icp < ncp; icp += per_device) {
	int i = icp / per_device;
        ProcessingState ps = ctx->pstates[i];

	ps->stored = 0;
    }

    return 0;
}

#ifdef DICT_SUPPORT_THREADS 
static int dictComputeThread(HWThread thr, void *hwctx, int device_id, void *data) {
    int err;
    DICTContext ctx = (DICTContext)data;
    ProcessingState ps = ctx->pstates[device_id];

    err = fftGetCurrentPoints(ps);
    ps->status = err;
    
    return 0;
}
#endif /* DICT_SUPPORT_THREADS */

int dictCompute(DICTContext ctx) {
    int err = 0;

    int icp, ncp = ctx->ncp;
    int per_device = ctx->ncp_per_device;
    
#ifdef DICT_HW_MEASURE_TIMINGS
    struct timeval tv1, tv2;
    gettimeofday(&tv1, NULL);
#endif /* DICT_HW_MEASURE_TIMINGS */
    
    if (ctx->use_threads) {
#ifdef DICT_SUPPORT_THREADS 
	hw_sched_schedule_task(ctx->sched, ctx, dictComputeThread);
	hw_sched_wait_task(ctx->sched);

	CHECK_ERROR(err)
#endif /* DICT_SUPPORT_THREADS */
    } else {
	for (icp = 0; icp < ncp; icp += per_device) {
	    int i = icp / per_device;
	    ProcessingState ps = ctx->pstates[i];
	    err = fftGetCurrentPoints(ps);
	    if (err) break;
	}
    }

#ifdef DICT_HW_MEASURE_TIMINGS
    gettimeofday(&tv2, NULL);
    ctx->time[16] += (tv2.tv_sec - tv1.tv_sec) * 1000000 + (tv2.tv_usec - tv1.tv_usec);
#endif /* DICT_HW_MEASURE_TIMINGS */
    
    return err;
}

int dictGetCurrentPoints(DICTContext ctx, float *res_x, float *res_y) {
    int err;
    int stored;

#ifdef DICT_HW_MEASURE_TIMINGS
    struct timeval tv1, tv2;
#endif /* DICT_HW_MEASURE_TIMINGS */

    err = dictCompute(ctx);
    if (err) return err;

#ifdef DICT_HW_MEASURE_TIMINGS
    gettimeofday(&tv1, NULL);
#endif /* DICT_HW_MEASURE_TIMINGS */
    
	// stored is changed synchronously across devices
    stored = ctx->pstates[0]->stored;

    if ((res_x)&&(res_x != ctx->res_x)) {
        float *data_x;
	if (stored) data_x = ctx->res_x;
	else data_x = ctx->points + 2 * ctx->ncp_alloc_size;
	
	memcpy(res_x, data_x, ctx->ncp * sizeof(float));
    }

    if ((res_y)&&(res_y != ctx->res_y)) {
	float *data_y;
        if (stored) data_y = ctx->res_y;
        else data_y = ctx->points + 3 * ctx->ncp_alloc_size;

	memcpy(res_y, data_y, ctx->ncp * sizeof(float));
    }

#ifdef DICT_HW_MEASURE_TIMINGS
    gettimeofday(&tv2, NULL);
    ctx->time[16] += (tv2.tv_sec - tv1.tv_sec) * 1000000 + (tv2.tv_usec - tv1.tv_usec);
#endif /* DICT_HW_MEASURE_TIMINGS */
    
    return 0;
}

/*
static int dictLoadTemplateFragment(DICTContext ctx, int icp, int ncp, const unsigned char *img) {
    int err;

    err = fftLoadBaseImage(ps, img);
    if (!err) err = fftLoadBaseFragment(ps, icp, min2(CP_BLOCK, ps->ncp - icp), img);

    return err;
}
*/


#ifdef DICT_SUPPORT_THREADS 
static int dictLoadTemplateImageThread(HWThread thr, void *hwctx, int device_id, void *data) {
    int err;
    DICTContext ctx = (DICTContext)data;
    ProcessingState ps = ctx->pstates[device_id];

    int jcp;

    err = fftLoadBaseImage(ps, ctx->param_img);
    
    if (ps->use_cache) {
	for (jcp = 0; (jcp < ps->ncp)&&(!err); jcp += CP_BLOCK) {
	    err = fftLoadBaseFragment(ps, jcp, min2(CP_BLOCK, ps->ncp - jcp), ctx->param_img);
	}
    } else if (!ps->base_mode) ctx->res_do_copy = 1;
	
    ps->status = err;

    return 0;
}
#endif /* DICT_SUPPORT_THREADS */

int dictLoadTemplateImage(DICTContext ctx, const unsigned char *img, int width, int height) {
    int err = 0;
    int jcp;
    int icp, ncp = ctx->ncp;
    int per_device = ctx->ncp_per_device;

    int do_copy = 0;
    
#ifdef DICT_HW_MEASURE_TIMINGS
    struct timeval tv1, tv2;
    gettimeofday(&tv1, NULL);
#endif /* DICT_HW_MEASURE_TIMINGS */

    err = dictSetDimensions(ctx, width, height);
    if (err) return err;

    if (ctx->use_threads) {
#ifdef DICT_SUPPORT_THREADS 
	ctx->res_do_copy = 0;
	ctx->param_img = img;
    
	hw_sched_schedule_task(ctx->sched, ctx, dictLoadTemplateImageThread);
	hw_sched_wait_task(ctx->sched);

	CHECK_ERROR(err)
    
	do_copy = ctx->res_do_copy;
#endif /* DICT_SUPPORT_THREADS */
    } else {
	for (icp = 0; icp < ncp; icp += per_device) {
	    int i = icp / per_device;
	    ProcessingState ps = ctx->pstates[i];

	    err = fftLoadBaseImage(ps, img);
    
	    if (ps->use_cache) {
		for (jcp = 0; (jcp < ps->ncp)&&(!err); jcp += CP_BLOCK) {
		    err = fftLoadBaseFragment(ps, jcp, min2(CP_BLOCK, ps->ncp - jcp), img);
		}
	    } else if (!ps->base_mode) do_copy = 1;
	
	    if (err) break;
	}
    }

    if (do_copy) {
	memcpy(ctx->base_image, img, width * height * sizeof(uint8_t));
    }

#ifdef DICT_HW_MEASURE_TIMINGS
    gettimeofday(&tv2, NULL);
    ctx->time[17] += (tv2.tv_sec - tv1.tv_sec) * 1000000 + (tv2.tv_usec - tv1.tv_usec);
#endif /* DICT_HW_MEASURE_TIMINGS */

    return err;
}


static int dictLoadFragment(DICTContext ctx, ProcessingState ps, int icp, int ncp, const unsigned char *input) {
    int err;
    int load_base = !ps->use_cache;

#ifdef DICT_HW_MEASURE_TIMINGS
    int time[8];
    struct timeval tv1, tv2;
    gettimeofday(&tv1, NULL);
#endif /* DICT_HW_MEASURE_TIMINGS */

    if (load_base) {
	err = fftLoadBaseFragment(ps, icp, ncp, ctx->base_image);
	if (err) return err;
    }

#ifdef DICT_HW_MEASURE_TIMINGS
    gettimeofday(&tv2, NULL);
    time[1] = (tv2.tv_sec - tv1.tv_sec) * 1000000 + (tv2.tv_usec - tv1.tv_usec);
#endif /* DICT_HW_MEASURE_TIMINGS */

    err = fftLoadFragment(ps, icp, ncp, input);
    if (err) return err;

#ifdef DICT_HW_MEASURE_TIMINGS
    gettimeofday(&tv2, NULL);
    time[2] = (tv2.tv_sec - tv1.tv_sec) * 1000000 + (tv2.tv_usec - tv1.tv_usec);
#endif /* DICT_HW_MEASURE_TIMINGS */

    err = fftPreprocessFragment(ps, icp, ncp);
    if (err) return err;

#ifdef DICT_HW_MEASURE_TIMINGS
    gettimeofday(&tv2, NULL);
    time[3] = (tv2.tv_sec - tv1.tv_sec) * 1000000 + (tv2.tv_usec - tv1.tv_usec);
#endif /* DICT_HW_MEASURE_TIMINGS */

    err = fftProcessFragment(ps, icp, ncp);
    if (err) return err;

#ifdef DICT_HW_MEASURE_TIMINGS
    gettimeofday(&tv2, NULL);
    time[4] = (tv2.tv_sec - tv1.tv_sec) * 1000000 + (tv2.tv_usec - tv1.tv_usec);
#endif /* DICT_HW_MEASURE_TIMINGS */

    err = fftPostprocessFragment(ps, icp, ncp);
    if (err) return err;

# ifdef DICT_HW_MEASURE_TIMINGS
    gettimeofday(&tv2, NULL);
    time[5] = (tv2.tv_sec - tv1.tv_sec) * 1000000 + (tv2.tv_usec - tv1.tv_usec);
    
    time[0] = 0;
    for (int i=5;i>0;i--) {
	time[i] -= time[i-1];
    }
    
    if (!ctx->use_threads) {
	for (int i = 1; i < 6; i++) {
	    ctx->time[i+5] += time[i];
	}
    }
# endif /* DICT_HW_MEASURE_TIMINGS */


    return 0;
}

#ifdef DICT_SUPPORT_THREADS
static int dictLoadImageThread(HWThread thr, void *hwctx, int device_id, void *data) {
    int err;
    DICTContext ctx = (DICTContext)data;
    ProcessingState ps = ctx->pstates[device_id];

    int jcp;

    err = fftLoadImage(ps, ctx->param_img);
    
    for (jcp = 0; (jcp < ps->ncp)&&(!err); jcp += CP_BLOCK) {
	err = dictLoadFragment(ctx, ps, jcp, min2(CP_BLOCK, ps->ncp - jcp), ctx->param_img);
    }

    ps->status = err;

    return 0;
}
#endif /* DICT_SUPPORT_THREADS */



int dictLoadImage(DICTContext ctx, const unsigned char *img) {
    int err = 0;

    int icp, ncp = ctx->ncp;
    int per_device = ctx->ncp_per_device;

    if (ctx->use_threads) {
#ifdef DICT_SUPPORT_THREADS
# ifdef DICT_HW_MEASURE_TIMINGS
	struct timeval tv1, tv2;
	gettimeofday(&tv1, NULL);
# endif /* DICT_HW_MEASURE_TIMINGS */

        ctx->param_img = img;
    
	hw_sched_schedule_task(ctx->sched, ctx, dictLoadImageThread);
	
	if (ctx->preload&DICT_PRELOAD_REQUEST) {
	    int preload_err = dictImageLoadImage(ctx, ctx->preload_request, ctx->image2);
	    if (preload_err) {
		reportError("Failed to preload image %s, error %i. Ignoring and continuing...", ctx->preload_request, preload_err);
		ctx->preload = 0;
	    } else {
		char *tmp_image_buf = ctx->preload_image;
		int tmp_image_size = ctx->preload_image_size;
		
		ctx->preload_image = ctx->preload_request;
		ctx->preload_image_size = ctx->preload_request_size;
		
		ctx->preload_request = tmp_image_buf;
		ctx->preload_request_size = tmp_image_size;
		
		ctx->preload = DICT_PRELOAD_READY;
	    }
	}
	
        hw_sched_wait_task(ctx->sched);

	CHECK_ERROR(err)
# ifdef DICT_HW_MEASURE_TIMINGS
	gettimeofday(&tv2, NULL);
	ctx->time[5] += (tv2.tv_sec - tv1.tv_sec) * 1000000 + (tv2.tv_usec - tv1.tv_usec);
# endif /* DICT_HW_MEASURE_TIMINGS */
#endif /* DICT_SUPPORT_THREADS */
    } else {
	for (icp = 0; icp < ncp; icp += per_device) {
    	    int i = icp / per_device;
	    ProcessingState ps = ctx->pstates[i];

#ifdef DICT_HW_MEASURE_TIMINGS
	    struct timeval tv1, tv2;
	    gettimeofday(&tv1, NULL);
#endif /* DICT_HW_MEASURE_TIMINGS */

	    err = fftLoadImage(ps, img);

#ifdef DICT_HW_MEASURE_TIMINGS
	    gettimeofday(&tv2, NULL);
	    ctx->time[4] += (tv2.tv_sec - tv1.tv_sec) * 1000000 + (tv2.tv_usec - tv1.tv_usec);
#endif /* DICT_HW_MEASURE_TIMINGS */

	    for (int jcp = 0; (jcp < ps->ncp)&&(!err); jcp += CP_BLOCK) {
		err = dictLoadFragment(ctx, ps, jcp, min2(CP_BLOCK, ps->ncp - jcp), img);
	    }
    
#ifdef DICT_HW_MEASURE_TIMINGS
	    gettimeofday(&tv2, NULL);
	    ctx->time[5] += (tv2.tv_sec - tv1.tv_sec) * 1000000 + (tv2.tv_usec - tv1.tv_usec);
#endif /* DICT_HW_MEASURE_TIMINGS */
	}
    }

    return err;
}

int dictPreloadImageFile(DICTContext ctx, const char *name) {
    size_t len = strlen(name) + 1;
    
    if (len > ctx->preload_request_size) {
	if (ctx->preload_request) free(ctx->preload_request);
	ctx->preload_request = (char*)malloc(2 * len * sizeof(char));
	if (!ctx->preload_request) {
	    ctx->preload &= ~DICT_PRELOAD_REQUEST;
	    ctx->preload_request_size = 0;
	    return DICT_ERROR_MALLOC;
	}
    }
    
    memcpy(ctx->preload_request, name, len);
    ctx->preload |= DICT_PRELOAD_REQUEST;

    return 0;
}

int dictLoadTemplateImageFile(DICTContext ctx, const char *name) {
    int err = dictImageLoadTemplateImage(ctx, name);
    if (err) return err;
    return dictLoadTemplateImage(ctx, ctx->image, ctx->width, ctx->height);
}

int dictLoadImageFile(DICTContext ctx, const char *name) {
    int err;
    unsigned char *tmp;

    if ((ctx->preload&DICT_PRELOAD_READY)&&(!strcmp(ctx->preload_image, name))) {
	tmp  = ctx->image2;
	ctx->image2 = ctx->image;
	ctx->image = tmp;
		
	ctx->preload &= ~DICT_PRELOAD_READY;
    } else {
    	err = dictImageLoadImage(ctx, name, NULL);
	if (err) return err;
    }
    
    return dictLoadImage(ctx, ctx->image);
}

int dictProcessImage(DICTContext ctx, const unsigned char *img) {
    int err = dictLoadImage(ctx, img);
    if (err) return err;
    return dictCompute(ctx);
}

int dictProcessImageFile(DICTContext ctx, const char *name) {
    int err = dictLoadImageFile(ctx, name);
    if (err) return err;
    return dictCompute(ctx);
}
