#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <sys/types.h>
#include <sys/stat.h>

#ifdef DICT_HW_MEASURE_TIMINGS
# include <sys/time.h>
#endif /* DICT_HW_MEASURE_TIMINGS */

#if defined(_WIN32) || defined(_WIN64)
# include "win/getopt.h"
#else 
# include <getopt.h>
#endif

#include <glib.h>

#include <dict_hw.h>

#include "helpers.h"

#define DATA_CLEANUP \
    if (fy) fclose(fy); \
    if (fx) fclose(fx); \
    free(res); \
    dictDestroyContext(ps); \
    free(images);


int matlab_mode = 0;

typedef enum {
    OPT_DEFAULT_FFT = 250,
    OPT_SINGLE_MODE = 251,
    OPT_NO_PRELOAD = 252,
    OPT_MULTIPASS = 253,
    OPT_CORRSIZE = 's',
    OPT_PRECISION = 'p',
    OPT_MATLAB = 'm',
    OPT_HELP = 'h'
} Options;

static struct option long_options[] = {
    {"matlab",			no_argument, 		0, OPT_MATLAB },
    {"corrsize",		required_argument,	0, OPT_CORRSIZE },	
    {"precision",		required_argument,	0, OPT_PRECISION },
    {"default-fft",		no_argument, 		0, OPT_DEFAULT_FFT },
    {"no-threads",		no_argument, 		0, OPT_SINGLE_MODE },
    {"no-preload",		no_argument, 		0, OPT_NO_PRELOAD },
    {"multipass",		no_argument, 		0, OPT_MULTIPASS },
    {"help",			no_argument, 		0, OPT_HELP },
    { 0, 0, 0, 0 }
};

void Usage(int argc, char *argv[]) {
    printf(
" %s [options] <directory> [maximages]\n"
"  Options:\n"
"	--matlab	- Matlab mode (transposed data ordering)\n"
"	--corrsize <s>	- Size of correlation area in pixels (15 by default)\n"
"	--precision <p>	- Subpixel precision (1000 by default)\n"
"	--default-fft	- Do not optimize FFT size\n"
"	--no-threads	- Disables threading, results in a single GPU mode\n"
"	--no-preload	- Disables image preloading in multi-threaded mode\n"
"	--multipass	- In multipass mode the caching is always used at the\n"
"			  expense of the repeated image reading\n"
"	--help		- Help message\n"
"\n\n", argv[0]);
}

int readfile(char *name, float *x, int size) {
    int i;

    char first_line[16384];

    struct stat st;
    stat(name, &st);

    FILE *f = fopen(name, "r");

    if (!f) {
	printf("File %s is not found\n", name);
	g_free(name);
	exit(-1);
    }

    g_free(name);


    if (matlab_mode) {
	int len = 0;
        int columns = 0;
	int block_len;
	do {
	    fgets(first_line, 16384, f);
	    for (block_len = 0; first_line[block_len]; block_len++) {
		if (first_line[block_len] == '.') columns++;
	    }
	    len += block_len;
	} while (block_len == 16383);
    
	int lines = st.st_size / len;
    
	int pos = 0;
	rewind(f);
	for (i = 0; fscanf(f, "%f", x+pos) == 1; i++) {
	    if ((pos+1) == size) {
		size *= 2;
		x = (float*)realloc(x, size);
	    }
	
	    pos = ((i+1)%columns)*lines + ((i+1)/columns);
	}
    } else {
	for (i = 0; fscanf(f, "%f", x+i) == 1; i++) {
	    if ((i+1) == size) {
		size *= 2;
		x = (float*)realloc(x, size);
	    }
	}
    }

    fclose(f);

    return i;
}

int writefile(int matlab_mode, char *name, float *x, int img, int ncp) {
    FILE *f = fopen(name, "w");
    if (!f) {
	printf("Can't write file %s\n", name);
	g_free(name);
	exit(-1);
    }
    g_free(name);

    if (matlab_mode) {
	for (int j = 0; j < ncp ; j++) {    
	    for (int i = 0; i < img ; i++) {
		fprintf(f, "% 9.7le\t", x[i*ncp + j]);
	    }
	    fprintf(f,"\n");
	}
    } else {
	for (int j = 0; j < img ; j++) {    
	    for (int i = 0; i < ncp ; i++) {
		fprintf(f, "% 9.7le\t", x[j*ncp + i]);
	    }
	    fprintf(f,"\n");
	}
    }
    
    fclose(f);

    return 0;
}

int compare(const void *a, const void*b) {
    return strcmp(*(char**)a, *(char**)b);
}

int main(int argc, char *argv[]) {
    int i;
    int err;
    int flags;
    
    int maximg, allocimg;
    
    int size = 32768;
    int xsize, ysize;
    float *x;
    float *y;

    char **images;

#ifdef DICT_HW_MEASURE_TIMINGS
    int time[16];
    struct timeval tv1, tv2, tv3;

    memset(time, 0, sizeof(time));
    gettimeofday(&tv1, NULL);
#endif /* DICT_HW_MEASURE_TIMINGS */

    int corrsize = 15;
    int precision = 1000;

    int default_fft = 0;
    int single_mode = 0;
    int preload = 1;

    int multipass = 0;
    
    int option_index = 0;
    unsigned char c;
    while ((c = getopt_long(argc, argv, "yhe:f:l:t:c:", long_options, &option_index)) != (unsigned char)-1) {
	switch (c) {
	    case 0:
	    break;
	    case OPT_MATLAB:
		matlab_mode = 1;
	    break;
	    case OPT_CORRSIZE:
		corrsize = atoi(optarg);
		if (corrsize < 1) {
		    printf("Invalid size of correlation are is specified, value \"%s\" is invalid", optarg);
		    exit(-1);
		}
	    break;
	    case OPT_PRECISION:
		precision = atoi(optarg);
		if (corrsize < 1) {
		    printf("Invalid subpixel precision is specified, value \"%s\" is invalid", optarg);
		    exit(-1);
		}
	    break;
	    case OPT_DEFAULT_FFT:
		default_fft = 1;
	    break;
	    case OPT_SINGLE_MODE:
		single_mode = 1;
		preload = 0;
	    break;
	    case OPT_NO_PRELOAD:
		preload = 0;
	    break;
	    case OPT_MULTIPASS:
		multipass = 1;
	    break;
	    case OPT_HELP:
		Usage(argc, argv);
		exit(0);		
	    break;
	    default:
		Usage(argc, argv);
		exit(0);
	}
    }

    if (optind > argc - 1) {
	Usage(argc, argv);
	exit(0);
    }
    
    x = (float*)malloc(size * sizeof(float));
    y = (float*)malloc(size * sizeof(float));

    gchar *name = g_strdup_printf("%s/data/grid_x.dat", argv[optind]);
    xsize = readfile(name, x, size);

    name = g_strdup_printf("%s/data/grid_y.dat", argv[optind]);
    ysize = readfile(name, y, size);
    
    if (xsize != ysize) {
	printf("Amount of X and Y coordinates of control points does not match\n");
	exit(-1);
    }
    
    //xsize = ysize = 1024;

    if (optind < (argc - 1)) maximg = atoi(argv[optind+1]);
    else maximg = 0;
    
    images = (char**)malloc(size * sizeof(char*));
    
    name = g_strdup_printf("%s/images", argv[optind]);
    GDir *dir = g_dir_open(name, 0, NULL);

    if (!dir) {
	printf("Directory %s is not found\n", name);
	g_free(name);
	exit(-1);
    }
    
    g_free(name);

    GRegex *regex = g_regex_new("(tiff?|bmp|jpe?g|png)", G_REGEX_CASELESS, (GRegexMatchFlags)0, NULL);
    if (!regex) {
	printf("Internal error. Failed to compile regular expression\n");
	exit(-1);
    }
    
    
    const gchar *file = g_dir_read_name(dir);
    for (i = 0;file;file = g_dir_read_name(dir)) {
	if (!g_regex_match(regex, file, (GRegexMatchFlags)0, NULL)) continue;
	
	if (i == size) {
	    size *= 2;
	    images = (char**)realloc(images, size);
	}
	images[i++] = g_strdup(file);
/*	if (!strstr(images[i-1], "tif")) {
	    puts(images[i-1]);
	    exit(-1);
	}*/
    }    
    g_dir_close(dir);
    g_regex_unref(regex);

    allocimg = i;
    if ((!maximg)||(maximg > i)) maximg = i;
    
    if (maximg < 2) {
	printf("At least two images are required\n");
	exit(-1);
    }

    qsort(images, allocimg, sizeof(char*), (int (*)(const void*, const void*))compare);

    int ncp = xsize;
    
    DICTContext ps = dictCreateContext();
    if (!ps) {
	printf("Context intialization is failed, no CUDA-devices...\n");
	exit(-1);
    }

    flags = DICT_FLAGS_DEFAULT;
    if (matlab_mode) flags |= DICT_FLAGS_MATLAB_MODE;
    if (default_fft) flags |= DICT_FLAGS_FIXED_FFT_SIZE;
    if (single_mode) flags |= DICT_FLAGS_SINGLE_THREAD;


    size_t ncp_block;
    
    if (multipass) {
	ncp_block = dictGetMaxCacheablePoints(corrsize, precision, flags);
	if (ncp_block >= ncp) multipass = 0;
    }

    err = dictSetup(ps, multipass?ncp_block:ncp, corrsize, precision, flags);
    if (err) {
	dictDestroyContext(ps);
	free(images);
	printf("Setup is failed, error code %i\n", err);
	exit(-1);
    }

    float *res;
    int step;

    FILE *fx = NULL, *fy = NULL;

    if ((matlab_mode)||(multipass)) {
	res = (float*) malloc(2 * ncp * (maximg - 1) * sizeof(float));
	step = ncp * (maximg - 1);
	
	if (!res) {
	    dictDestroyContext(ps);
	    free(images);
	    printf("Can't allocate result memory\n");
	    exit(-1);
	}
    } else {
	res = (float*) malloc(2 * ncp * sizeof(float));
	step = ncp;
	dictSetPointsBuffer(ps, res, res + ncp);

	name = g_strdup_printf("%s/data/validx.dat", argv[optind]);
	fx = fopen(name, "w");
	g_free(name);
	name = g_strdup_printf("%s/data/validy.dat", argv[optind]);
	fy = fopen(name, "w");
	g_free(name);
	
	if ((!res)||(!fx)||(!fy)) {
	    if (fy) fclose(fy);
	    if (fx) fclose(fx);
	    if (res) free(res);
	    dictDestroyContext(ps);
	    free(images);
	    printf("Can't create output files...\n");
	    exit(-1);
	}
    }

#ifdef DICT_HW_MEASURE_TIMINGS
	gettimeofday(&tv2, NULL);
	time[0] = (tv2.tv_sec - tv1.tv_sec) * 1000000 + (tv2.tv_usec - tv1.tv_usec);
#endif /* DICT_HW_MEASURE_TIMINGS */

    size_t icp;
    size_t real_ncp = ncp;
    
    for (icp = 0, ncp = multipass?ncp_block:ncp; icp < real_ncp; icp += ncp, ncp = min2(ncp_block, real_ncp - icp)) {
#ifndef DICT_HW_MEASURE_TIMINGS
	if (multipass) {
	    printf("Processing control points %i + %i (of %i)\n", icp, ncp, real_ncp);
	}
#endif /* ! DICT_HW_MEASURE_TIMINGS */

	if (icp) {
#ifdef DICT_HW_MEASURE_TIMINGS
	    gettimeofday(&tv3, NULL);
#endif /* DICT_HW_MEASURE_TIMINGS */
	    dictSetActualNumberOfPoints(ps, ncp);
	    dictClean(ps);
	    //dictSetup(ps, ncp, corrsize, precision, flags);

#ifdef DICT_HW_MEASURE_TIMINGS
	    gettimeofday(&tv2, NULL);
	    time[6] += (tv2.tv_sec - tv3.tv_sec) * 1000000 + (tv2.tv_usec - tv3.tv_usec);
#endif /* DICT_HW_MEASURE_TIMINGS */
	}

#ifdef DICT_HW_MEASURE_TIMINGS
	gettimeofday(&tv3, NULL);
#endif /* DICT_HW_MEASURE_TIMINGS */
	dictSetTemplatePoints(ps, x + icp, y + icp);
	dictSetCurrentPoints(ps, x + icp, y + icp);
#ifdef DICT_HW_MEASURE_TIMINGS
	gettimeofday(&tv2, NULL);
        time[7] += (tv2.tv_sec - tv3.tv_sec) * 1000000 + (tv2.tv_usec - tv3.tv_usec);
#endif /* DICT_HW_MEASURE_TIMINGS */

#ifdef DICT_HW_MEASURE_TIMINGS
	gettimeofday(&tv3, NULL);
#endif /* DICT_HW_MEASURE_TIMINGS */
	name = g_strdup_printf("%s/images/%s", argv[optind], images[0]);
	if (!g_file_test(name, G_FILE_TEST_IS_REGULAR)) {
	    g_free(name);
	    printf("%s is not a regular file\n", images[0]);
	    DATA_CLEANUP
	    exit(-1);
	}
    
	err = dictLoadTemplateImageFile(ps, name);
	g_free(name);
	if (err) {
	    printf("failed to load template file %s\n", images[0]);
	    DATA_CLEANUP
	    exit(-1);
	}
#ifdef DICT_HW_MEASURE_TIMINGS
	gettimeofday(&tv2, NULL);
        time[8] += (tv2.tv_sec - tv3.tv_sec) * 1000000 + (tv2.tv_usec - tv3.tv_usec);
#endif /* DICT_HW_MEASURE_TIMINGS */
    

#ifdef DICT_HW_MEASURE_TIMINGS
        gettimeofday(&tv3, NULL);
#endif /* DICT_HW_MEASURE_TIMINGS */
	gchar *next = NULL;

        name = g_strdup_printf("%s/images/%s", argv[optind], images[1]);
	if (!g_file_test(name, G_FILE_TEST_IS_REGULAR)) {
	    g_free(name);
	    printf("%s is not a regular file\n", images[1]);
	    DATA_CLEANUP
	    exit(-1);
	}

#ifdef DICT_HW_MEASURE_TIMINGS
	gettimeofday(&tv2, NULL);
	time[5] += (tv2.tv_sec - tv3.tv_sec) * 1000000 + (tv2.tv_usec - tv3.tv_usec);
#endif /* DICT_HW_MEASURE_TIMINGS */
    
	for (i = 1; i < maximg; i++, name = next) {
#ifndef DICT_HW_MEASURE_TIMINGS
	    printf("%s\n", images[i]);
#endif /* !DICT_HW_MEASURE_TIMINGS */

#ifdef DICT_HW_MEASURE_TIMINGS
	    gettimeofday(&tv3, NULL);
#endif /* DICT_HW_MEASURE_TIMINGS */
	    if ((i + 1) < maximg) {
		next = g_strdup_printf("%s/images/%s", argv[optind], images[i + 1]);
	        if (!g_file_test(next, G_FILE_TEST_IS_REGULAR)) {
		    g_free(next);
		    g_free(name);
		    printf("%s is not a regular file\n", images[i + 1]);
		    DATA_CLEANUP
		    exit(-1);
		}
	    
	        if (preload) {
	    	    dictPreloadImageFile(ps, next);
		}
	    }
	
	    err = dictLoadImageFile(ps, name);
#ifdef DICT_HW_MEASURE_TIMINGS
	    gettimeofday(&tv2, NULL);
	    time[5] += (tv2.tv_sec - tv3.tv_sec) * 1000000 + (tv2.tv_usec - tv3.tv_usec);
#endif /* DICT_HW_MEASURE_TIMINGS */
	
	    g_free(name);

	    if (err) {
		printf("failed to load image %s\n", images[0]);
		DATA_CLEANUP
	        exit(-1);
	    }
	

	    if ((matlab_mode)||(multipass)) {
		float *out = res + real_ncp * (i - 1) + icp;
		dictGetCurrentPoints(ps, out, out + step);
	    } else {
		dictCompute(ps);

#ifdef DICT_HW_MEASURE_TIMINGS
		gettimeofday(&tv3, NULL);
#endif /* DICT_HW_MEASURE_TIMINGS */

	        for (int j = 0; j < ncp; j++) {
		    fprintf(fx, "% 9.7le\t", res[j]);
		}
	        fprintf(fx,"\n");
	    
		float *out = res + step;
	        for (int j = 0; j < ncp; j++) {
		    fprintf(fy, "% 9.7le\t", out[j]);
		}
		fprintf(fy,"\n");

#ifdef DICT_HW_MEASURE_TIMINGS
		gettimeofday(&tv2, NULL);
	        time[1] += (tv2.tv_sec - tv3.tv_sec) * 1000000 + (tv2.tv_usec - tv3.tv_usec);
#endif /* DICT_HW_MEASURE_TIMINGS */
	    }
	}
    }
    
#ifdef DICT_HW_MEASURE_TIMINGS
    gettimeofday(&tv3, NULL);
#endif /* DICT_HW_MEASURE_TIMINGS */
    
    if ((matlab_mode)||(multipass)) {
	name = g_strdup_printf("%s/data/validx.dat", argv[optind]);
	writefile(matlab_mode, name, res, maximg - 1, real_ncp);

	name = g_strdup_printf("%s/data/validy.dat", argv[optind]);
	writefile(matlab_mode, name, res+step, maximg - 1, real_ncp);
    } else {
	fclose(fx);
	fclose(fy);
    }

#ifdef DICT_HW_MEASURE_TIMINGS
    gettimeofday(&tv2, NULL);
    time[1] += (tv2.tv_sec - tv3.tv_sec) * 1000000 + (tv2.tv_usec - tv3.tv_usec);
    time[2] = (tv2.tv_sec - tv1.tv_sec) * 1000000 + (tv2.tv_usec - tv1.tv_usec) - time[0];
#endif /* DICT_HW_MEASURE_TIMINGS */

    for (i = 0; i < allocimg; i++) {
	g_free(images[i]);
    }
    
    dictDestroyContext(ps);
    
    free(res);
    free(images);        
    free(y);
    free(x);

#ifdef DICT_HW_MEASURE_TIMINGS
    gettimeofday(&tv3, NULL);
    time[3] = (tv3.tv_sec - tv1.tv_sec) * 1000000 + (tv3.tv_usec - tv1.tv_usec);
    time[4] = (tv3.tv_sec - tv2.tv_sec) * 1000000 + (tv3.tv_usec - tv2.tv_usec);
    
    time[9] = time[6] + time[7] + time[8];
    
    reportMessage("");
    reportMessage("Global timings");
    reportMessage("--------------");
    print_timing("Overall", time[3]);
    print_timing("Initialization", time[0] + time[9]);
    print_timing("Processing/Full", time[2] - time[9], maximg - 1);
    print_timing("Cleanup", time[4]);
    print_timing("Initialization/Reinit", time[6]);
    print_timing("Initialization/SetPoints", time[7]);
    print_timing("Initialization/LoadTemplate", time[8]);
    print_timing("Processing/Computations", time[2] - time[9] - time[1], maximg - 1);
    print_timing("Processing/Writeout", time[1], maximg - 1);
    print_timing("Computations/Load Image", time[5], maximg - 1);
    print_timing("Computations/Waiting Result", time[2] - time[9] - time[1] - time[5], maximg - 1);
    reportMessage("");
#endif /* DICT_HW_MEASURE_TIMINGS */

}

